from .effects import MONO_EFFECTS, PRESET_EFFECTS, CUSTOM_EFFECTS
from .constants import MonoEffect, CustomEffect
from .WifiLedShopLight import WifiLedShopLight

__all__ = [
  'WifiLedShopLight',
  'MonoEffect',
  'CustomEffect',
  'DEFAULT_EFFECT',
  'MONO_EFFECTS',
  'PRESET_EFFECTS',
  'CUSTOM_EFFECTS',
]



from .constants import (MonoEffect, CustomEffect)

MONO_EFFECTS = {
  'Solid (custom color)': MonoEffect.SOLID,
  'Breathing (custom color)': MonoEffect.BREATHING,
  'Meteor (custom color)': MonoEffect.METEOR,
  'Flow (custom color)': MonoEffect.FLOW,
  'Wave (custom color)': MonoEffect.WAVE,
  'Flash (custom color)': MonoEffect.FLASH,
  'Stack (custom color)': MonoEffect.STACK,
  'Catchup (custom color)': MonoEffect.CATCHUP,
}

PRESET_EFFECTS = {
  'Rainbow': 0,
  'Colors changing meteors': 1,
  'Flowing Colors': 2,
  'Overlaying Flowing': 3,
  'Red stacking': 4,
  'Green stacking': 5,
  'Blue stacking': 6,
  'Colors changing stacking': 7,
  'Fading flowing': 8,
  'Color changing cross flowing': 9,
  'Cross flowing': 10,
  'Chasing color dots': 11,
  'Colorful wave': 12,
  'Burning fire': 13,
  'Red meteor': 14,
  'Green meteor': 15,
  'Blue meteor': 16,
  'Yellow meteor': 17,
  'Cyan meteor': 18,
  'Purple meteor': 19,
  'White meteor': 20,
  'Red wave': 21,
  'Green wave': 22,
  'Blue wave': 23,
  'Yellow wave': 24,
  'Cyan wave': 25,
  'Purple wave': 26,
  'White wave': 27,
  'Red chasing dots': 28,
  'Green chasing dots': 29,
  'Blue chasing dots': 30,
  'Yellow chasing dots': 31,
  'Cyan chasing dots': 32,
  'White chasing dots': 33,
  'Purple chasing dots': 34,
  'Cyan dots blink on silver': 35,
  'Purple dots blink on silver': 36,
  'Yellow dots blink on silver': 37,
  'Blue dots blink on silver': 38,
  'Green dots blink on silver': 39,
  'Red dots blink on silver': 40,
  'Red and Green flowing': 41,
  'Red and blue flowing': 42,
  'Red and yellow flowing': 43,
  'Red and cyan flowing': 44,
  'Red and purple flowing': 45,
  'Red and white flowing': 46,
  'Green and blue flowing': 47,
  'Green and yellow flowing': 48,
  'Green and cyan flowing': 49,
  'Green and purple flowing': 50,
  'Green and white flowing': 51,
  'Blue and yellow flowing': 52,
  'Blue and cyan flowing': 53,
  'Blue and purple flowing': 54,
  'Blue and white flowing': 55,
  'Yellow and cyan flowing': 56,
  'Yellow and purple flowing': 57,
  'Yellow and white flowing': 58,
  'Cyan and purple flowing': 59,
  'Cyan and white flowing': 60,
  'Red crossing flowing': 61,
  'Green crossing flowing': 62,
  'Blue crossing flowing': 63,
  'Yellow crossing flowing': 64,
  'Cyan crossing flowing': 65,
  'Purple crossing flowing': 66,
  'White crossing flowing': 67,
  'Red arrows': 68,
  'Green arrows': 69,
  'Blue arrows': 70,
  'Yellow arrows': 71,
  'Cyan arrows': 72,
  'Purple arrows': 73,
  'Red cudgel': 74,
  'Green cudgel': 75,
  'Blue cudgel': 76,
  'Yellow cudgel': 77,
  'Cyan cudgel': 78,
  'Purple cudgel': 79,
  'Gaps in cyan': 80,
  'Gaps in purple': 81,
  'Gaps in yellow': 82,
  'Gaps in blue': 83,
  'Gaps in green': 84,
  'Gaps in red': 85,
  'Red breathe': 86,
  'Green breathe': 87,
  'Blue breathe': 88,
  'Yellow breathe': 89,
  'Cyan breathe': 90,
  'Purple breathe': 91,
  'White breathe': 92,
  'Red arrows reverse': 93,
  'Green arrows reverse': 94,
  'Blue arrows reverse': 95,
  'Yellow arrows reverse': 96,
  'Cyan arrows reverse': 97,
  'Purple arrows reverse': 98,
  'Red dots running in cyan': 99,
  'Green dots running in purple': 100,
  'Blue dots running in yellow': 101,
  'Yellow dots running in blue': 102,
  'Cyan dots running in green': 103,
  'Purple dots running in red': 104,
  'White dots running in cyan': 105,
  'Red fading in and out': 106,
  'Green fading in and out': 107,
  'Blue fading in and out': 108,
  'Yellow fading in and out': 109,
  'Cyan fading in and out': 110,
  'Purple fading in and out': 111,
  'White fading in and out': 112,
  'Red meteor reverse': 113,
  'Green meteor reverse': 114,
  'Blue meteor reverse': 115,
  'Yellow meteor reverse': 116,
  'Cyan meteor reverse': 117,
  'Purple meteor reverse': 118,
  'White meteor reverse': 119,
  'Red wave 2': 120,
  'Green wave 2': 121,
  'Blue wave 2': 122,
  'Yellow wave 2': 123,
  'Cyan wave 2': 124,
  'Purple wave 2': 125,
  'White wave 2': 126,
  'Red wing dots': 127,
  'Green swing dots': 128,
  'Blue swing dots': 129,
  'Yellow swing dots': 130,
  'Cyan swing dots': 131,
  'Purple swing dots': 132,
  'White swing dots': 133,
  'Red and green cudgel': 134,
  'Green and blue cudgel': 135,
  'Blue and yellow cudgel': 136,
  'Yellow and cyan cudgel': 137,
  'Cyan and purple cudgel': 138,
  'Purple and white cudgel': 139,
  'White and red cudgel': 140,
  'Red overlaps green': 141,
  'Green overlaps red': 142,
  'Blue overlaps green': 143,
  'Yellow overlaps green': 144,
  'Cyan overlaps green': 145,
  'Purple overlaps green': 146,
  'White overlaps green': 147,
  'Red overlaps blue': 148,
  'Green overlaps blue': 149,
  'Blue overlaps green 2': 150,
  'Yellow overlaps blue': 151,
  'Cyan overlaps blue': 152,
  'Purple overlaps blue': 153,
  'White overlaps blue': 154,
  'Pink mix blue': 155,
  'Green mix yellow': 156,
  'Blue mix pink': 157,
  'Blue mix white': 158,
  'Green mix orange': 159,
  'Blue mix purple': 160,
  'Cyan mix white': 161,
  'Red blinking': 162,
  'Green blinking': 163,
  'Blue blinking': 164,
  'Yellow blinking': 165,
  'Cyan blinking': 166,
  'Purple blinking': 167,
  'White blinking': 168,
  'Red stacking into green': 169,
  'Green stacking into blue': 170,
  'Blue stacking into yellow': 171,
  'Yellow stacking into cyan': 172,
  'Cyan stacking into purple': 173,
  'Purple stacking into white': 174,
  'White stacking into red': 175,
  'Color changing breathing': 176,
  'Multicolored gradients': 177,
  'Color Flashing': 178,
  'Rainbow reverse': 179
}

CUSTOM_EFFECTS = {
  'Custom Effect 1': CustomEffect.CUSTOM_1,
  'Custom Effect 2': CustomEffect.CUSTOM_2,
  'Custom Effect 3': CustomEffect.CUSTOM_3,
  'Custom Effect 4': CustomEffect.CUSTOM_4,
  'Custom Effect 5': CustomEffect.CUSTOM_5,
  'Custom Effect 6': CustomEffect.CUSTOM_6,
  'Custom Effect 7': CustomEffect.CUSTOM_7,
  'Custom Effect 8': CustomEffect.CUSTOM_8,
  'Custom Effect 9': CustomEffect.CUSTOM_9,
  'Custom Effect 10': CustomEffect.CUSTOM_10,
  'Custom Effect 11': CustomEffect.CUSTOM_11,
  'Custom Effect 12': CustomEffect.CUSTOM_12,
}



import socket

from .constants import (Command, CommandFlag, MonoEffect)
from .utils import clamp
from .WifiLedShopLightState import WifiLedShopLightState

class WifiLedShopLight:
  """
  A Wifi LED Shop Light
  """

  def __init__(self, ip, port = 8189, timeout = 5, retries = 5):
    """
    Creates a new Wifi LED Shop light

    :param ip: The IP of the controller on the network (STA Mode, not AP mode).
    :param port: The port the controller should listen on. It should almost always be left as the default.
    :param timeout: The timeout in seconds to wait listening to the socket.
    :param retries: The number of times to retry sending a command if it fails or times out before giving up.
    """
    self.ip = ip
    self.port = port
    self.timeout = timeout
    self.retries = retries
    self.state = WifiLedShopLightState()

    self.sock = None
    self.reconnect()

  def __enter__(self):
    return self

  def __exit__(self, type, value, traceback):
    self.close()

  def reconnect(self):
    """
    Try to (re-)connect to the controller via a socket
    """
    if self.sock:
      self.close()

    self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    self.sock.settimeout(self.timeout)
    self.sock.connect((self.ip, self.port))

  def close(self):
    """
    Closes the socket connection to the light
    """
    self.sock.close()
    self.sock = None

  def set_color(self, r=0, g=0, b=0):
    """
    Sets the color of the light (rgb each 0 to 255)
    """
    r = clamp(r)
    g = clamp(g)
    b = clamp(b)
    self.state.color = (r, g, b)
    self.send_command(Command.SET_COLOR, [int(r), int(g), int(b)])

  def set_brightness(self, brightness=0):
    """
    Sets the brightness of the light

    :param brightness: An int describing the brightness (0 to 255, where 255 is the brightest)
    """
    brightness = clamp(brightness)
    self.state.brightness = brightness
    self.send_command(Command.SET_BRIGHTNESS, [int(brightness)])

  def set_speed(self, speed=0):
    """
    Sets the speed of the effect. Not all effects use the speed, but it can be safely set regardless
    
    :param speed: An int describing the speed an effect will play at. (0 to 255, where 255 is the fastest)
    """
    speed = clamp(speed)
    self.state.speed = speed
    self.send_command(Command.SET_SPEED, [int(speed)])

  def set_preset(self, preset=0):
    """
    Sets the light effect to the provided built-in effect number

    :param preset: The preset effect to use. Valid values are 0 to 255. See the MonoEffect enum, or MONO_EFFECTS and PRESET_EFFECTS for mapping.
    """
    preset = clamp(preset)
    self.state.mode = preset
    self.send_command(Command.SET_PRESET, [int(preset)])

  def set_custom(self, custom):
    """
    Sets the light effect to the provided custom effect number
    
    :param custom: The custom effect to use. Valid values are 1 to 12. See the CustomEffect enum.
    """
    custom = clamp(custom, 1, 12)
    self.state.mode = custom
    self.send_command(Command.SET_CUSTOM, [int(custom)])

  def toggle(self):
    """
    Toggles the state of the light without checking the current state
    """
    self.state.is_on = not self.state.is_on
    self.send_command(Command.TOGGLE)

  def turn_on(self):
    """
    Toggles the light on only if it is not already on
    """
    if not self.state.is_on:
      self.toggle()

  def turn_off(self):
    """
    Toggles the light off only if it is not already off
    """
    if self.state.is_on:
      self.toggle()

  def set_segments(self, segments):
    """
    Sets the total number of segments. Total lights is segments * lights_per_segment.
    
    :param segments: The number of segments
    """
    self.send_command(Command.SET_SEGMENT_COUNT, [segments])

  def set_lights_per_segment(self, lights_per_segment):
    """
    Sets the number of lights per segment. Total lights is segments * lights_per_segment.

    :param lights_per_segment: The number of lights per segment
    """
    lights_per_segment_data = list(lights_per_segment.to_bytes(2, byteorder='little'))
    self.send_command(Command.SET_LIGHTS_PER_SEGMENT, lights_per_segment_data)

  def set_calculated_segments(self, total_lights, segments):
    """
    Helper function to automatically set the number of segments and lights per segment
    to reach the target total lights (rounded down to never exceed total_lights)

    Usually you know the total number of lights you have available on a light strip
    and want to split it into segments that take up the whole strip

    :param total_lights: The target total number of lights to use
    :param segments: The number of segments to split the total into
    """
    self.set_segments(segments)
    self.set_lights_per_segment(int(total_lights / segments))

  def send_command(self, command, data=[]):
    """
    Helper method to send a command to the controller.

    Mostly for internal use, prefer the specific functions where possible.
    
    Formats the low level message details like Start/End flag, binary data, and command

    :param command: The command to send to the controller. See the Command enum for valid commands.
    """
    min_data_len = 3
    padded_data = data + [0] * (min_data_len - len(data))
    raw_data = [CommandFlag.START, *padded_data, command, CommandFlag.END]
    self.send_bytes(raw_data)

  def send_bytes(self, data):
    """
    Helper method to send raw bytes directly to the controller

    Mostly for internal use, prefer the specific functions where possible
    """
    raw_data = bytes(data)

    attempts = 0
    while True:
      try:
        self.sock.sendall(raw_data)
        return
      except (socket.timeout, BrokenPipeError):
        if (attempts < self.retries):
          self.reconnect()
          attempts += 1
        else:
          raise

  def sync_state(self):
    """
    Syncs the state of the controller with the state of this object
    """
    attempts = 0
    while True:
      try:
        # Send the request for sync data
        self.send_command(Command.SYNC)

        response = self.sock.recv(1024)

        # Extract the state data
        state = bytearray(response)
        self.state.update_from_sync(state)
        return
      except (socket.timeout, BrokenPipeError):
        # When there is an error with the socket, close the connection and connect again
        if attempts < self.retries:
          self.reconnect()
          attempts += 1
        else:
          raise

  def __repr__(self):
    return f"""WikiLedShopLight @ {self.ip}:{self.port}
      state: {self.state}
    """



from enum import IntEnum

class CommandFlag(IntEnum):
  """
  All messages to the device must be wrapped in a start and end flag

  Mostly for internal use, prefer to use the functions on WifiLedShopLight instead.
  """
  START = 0x38
  END = 0x83

class Command(IntEnum):
  """
  Commands that can be sent to the device.
  
  Mostly for internal use, prefer to use the functions on WifiLedShopLight instead.

  To be used with WifiLedShopLight.send_command()
  """
  TOGGLE = 0xAA
  SET_COLOR = 0x22
  SET_BRIGHTNESS = 0x2A
  SET_SPEED = 0x03
  SET_PRESET = 0x2C
  SET_CUSTOM = 0x02
  SET_LIGHTS_PER_SEGMENT = 0x2D
  SET_SEGMENT_COUNT = 0x2E
  SYNC = 0x10

class StatePosition(IntEnum):
  """
  State is returned from the device as a byte array. 
  
  This provides the position of each piece of useful data within the state bytearray.
  """
  IS_ON = 1
  COLOR_R = 10
  COLOR_G = 11
  COLOR_B = 12
  MODE = 2
  SPEED = 3
  BRIGHTNESS = 4

class MonoEffect(IntEnum):
  """
  Preset effects that are for a single, user customizable color

  To be used with WifiLedShopLight.set_preset()
  """
  SOLID = 211,
  BREATHING = 206,
  METEOR = 205,
  FLOW = 208,
  WAVE = 209,
  FLASH = 210,
  STACK = 207,
  CATCHUP = 212,

class CustomEffect(IntEnum):
  """
  Sets the brightness of the light (0 to 255)

  To be used with WifiLedShopLight.set_custom()
  """
  CUSTOM_1 = 1,
  CUSTOM_2 = 2,
  CUSTOM_3 = 3,
  CUSTOM_4 = 4,
  CUSTOM_5 = 5,
  CUSTOM_6 = 6,
  CUSTOM_7 = 7,
  CUSTOM_8 = 8,
  CUSTOM_9 = 9,
  CUSTOM_10 = 10,
  CUSTOM_11 = 11,
  CUSTOM_12 = 12,



from setuptools import setup

setup(
    name = 'pyledshop',
    packages = ['pyledshop'],
    version = '0.0.1',
    description = 'A Python library to communicate with LED Shop compatible led strip controllers',
    author = 'blockarchitech',
    url = 'https://github.com/blockarchitech/pyledshop',
    license = 'MIT',
    keywords = [
        'led shop',
        'led strip',
        'light',
        'SP108E',
        ],
		classifiers=[
			    'Development Status :: 4 - Beta',
		]
)



def clamp(value, min=0, max=255):
  """
  Clamps a value to be within the specified range

  Since led shop uses bytes for most data, the defaults are 0-255
  """
  if value > max:
    return max
  elif value < min:
    return min
  else:
    return value



from .constants import (StatePosition, MonoEffect)

class WifiLedShopLightState:
  def __init__(self):
    self.is_on = False
    self.color = (255, 255, 255)
    self.brightness = 255
    self.mode = MonoEffect.SOLID
    self.speed = 255

  def __repr__(self):
      return f"""
        is_on: {self.is_on}
        color: {self.color}
        brightness: {self.brightness}
        mode: {self.mode}
        speed: {self.speed}
      """

  def update_from_sync(self, sync_data):
    self.is_on = sync_data[StatePosition.IS_ON] == 1
    self.color = (
      sync_data[StatePosition.COLOR_R],
      sync_data[StatePosition.COLOR_G],
      sync_data[StatePosition.COLOR_B]
    )
    self.mode = sync_data[StatePosition.MODE]
    self.speed = sync_data[StatePosition.SPEED]
    self.brightness = sync_data[StatePosition.BRIGHTNESS]
